@@ -23,14 +23,26 @@
 
 #include "pkcs11/pkcs11.h"
 #include "pkcs11/pkcs11g.h"
+#include "pkcs11/pkcs11i.h"
 
+#include "gck-aes-key.h"
+#include "gck-aes-mechanism.h"
 #include "gck-attributes.h"
+#include "gck-certificate.h"
+#include "gck-credential.h"
 #include "gck-factory.h"
 #include "gck-manager.h"
+#include "gck-memory-store.h"
 #include "gck-module.h"
-#include "gck-private-key.h"
-#include "gck-public-key.h"
+#include "gck-null-key.h"
+#include "gck-null-mechanism.h"
+#include "gck-dh-private-key.h"
+#include "gck-private-xsa-key.h"
+#include "gck-dh-public-key.h"
+#include "gck-public-xsa-key.h"
 #include "gck-session.h"
+#include "gck-store.h"
+#include "gck-timer.h"
 #include "gck-transaction.h"
 #include "gck-util.h"
 
@@ -38,24 +50,40 @@ enum {
 	PROP_0,
 	PROP_MANAGER,
 	PROP_WRITE_PROTECTED,
-	PROP_INITIALIZE_ARGS
+	PROP_INITIALIZE_ARGS,
+	PROP_MUTEX
 };
 
+#define APARTMENT_APP(apt) \
+	((apt) & ~CK_GNOME_MAX_SLOT)
+#define APARTMENT_SLOT(apt) \
+	((apt) & CK_GNOME_MAX_SLOT)
+#define APARTMENT_ID(slot, app) \
+	(((slot) & CK_GNOME_MAX_SLOT) | ((app) & ~CK_GNOME_MAX_SLOT))
+
 struct _GckModulePrivate {
-	GckManager *token_manager; 
-	GHashTable *virtual_slots_by_id;        /* Various slot partitions by their ID */
+	GMutex *mutex;                          /* The mutex controlling entry to this module */
+
+	GckManager *token_manager;
+	GHashTable *apartments_by_id;           /* Apartment (slot + application) by their id */
 	GHashTable *sessions_by_handle;         /* Mapping of handle to all open sessions */
-	gint handle_counter;                    /* Constantly incrementing counter for handles and the like */
+	gulong handle_counter;                  /* Constantly incrementing counter for handles and the like */
 	GArray *factories;                      /* Various registered object factories */
 	gboolean factories_sorted;              /* Whether we need to sort the object factories */
+
+	GHashTable *transient_objects;          /* Token objects that are not stored permanently. */
+	GckStore *transient_store;              /* Store for trantsient objects. */
 };
 
-typedef struct _VirtualSlot {
+typedef struct _Apartment {
+	CK_ULONG apt_id;
 	CK_SLOT_ID slot_id;
+	CK_G_APPLICATION_ID app_id;
+	CK_G_APPLICATION_PTR app_ptr;
 	GckManager *session_manager;
 	GList *sessions;
-	gboolean logged_in;
-} VirtualSlot;
+	CK_USER_TYPE logged_in;
+} Apartment;
 
 /* Our slot identifier is 1 */
 #define GCK_SLOT_ID  1
@@ -67,7 +95,7 @@ G_DEFINE_TYPE (GckModule, gck_module, G_TYPE_OBJECT);
 static const CK_INFO default_module_info = {
 	{ CRYPTOKI_VERSION_MAJOR, CRYPTOKI_VERSION_MINOR },
 	"Gnome Keyring",
-	CKF_GNOME_APPARTMENTS,
+	CKF_G_APPLICATIONS,
 	"Gnome Keyring Module",
 	{ 1, 1 },
 };
@@ -123,9 +151,40 @@ static const MechanismAndInfo mechanism_list[] = {
 	 * CKM_DSA
 	 * For DSA, min and max are the minimum and maximum modulus in bits
 	 */
-	{ CKM_DSA, { 512, 1024, CKF_SIGN | CKF_VERIFY } }
+	{ CKM_DSA, { 512, 1024, CKF_SIGN | CKF_VERIFY } },
+
+	/*
+	 * CKM_DH_PKCS_KEY_PAIR_GEN
+	 * For DH derivation the min and max are sizes of prime in bits.
+	 */
+	{ CKM_DH_PKCS_KEY_PAIR_GEN, { 768, 8192, CKF_GENERATE_KEY_PAIR } },
+
+	/*
+	 * CKM_DH_PKCS_DERIVE
+	 * For DH derivation the min and max are sizes of prime in bits.
+	 */
+	{ CKM_DH_PKCS_DERIVE, { 768, 8192, CKF_DERIVE } },
+
+	/*
+	 * CKM_AES_CBC_PAD
+	 * For AES the min and max are sizes of key in bytes.
+	 */
+	{ CKM_AES_CBC_PAD, { GCK_AES_MECHANISM_MIN_LENGTH, GCK_AES_MECHANISM_MAX_LENGTH, CKF_WRAP | CKF_UNWRAP } },
+
+	/*
+	 * CKM_G_NULL
+	 * For NULL min and max are zero
+	 */
+	{ CKM_G_NULL, { GCK_NULL_MECHANISM_MIN_LENGTH, GCK_NULL_MECHANISM_MAX_LENGTH, CKF_WRAP | CKF_UNWRAP } },
 };
 
+/* Hidden function that you should not use */
+GMutex* _gck_module_get_scary_mutex_that_you_should_not_touch (GckModule *self);
+
+static void  remove_transient_object (GckModule *self, GckTransaction *transaction, GckObject *object);
+
+static void  add_transient_object    (GckModule *self, GckTransaction *transaction, GckObject *object);
+
 /* -----------------------------------------------------------------------------
  * INTERNAL 
  */
@@ -133,8 +192,8 @@ static const MechanismAndInfo mechanism_list[] = {
 static gint
 sort_factory_by_n_attrs (gconstpointer a, gconstpointer b)
 {
-	const GckFactoryInfo *fa = a;
-	const GckFactoryInfo *fb = b;
+	const GckFactory *fa = a;
+	const GckFactory *fb = b;
 	
 	g_assert (a);
 	g_assert (b);
@@ -158,76 +217,102 @@ extend_space_string (CK_UTF8CHAR_PTR string, gsize length)
 }
 
 static void
-virtual_slot_free (gpointer data)
+apartment_free (gpointer data)
 {
-	VirtualSlot *slot;
+	Apartment *apt;
 	GList *l;
-	
+
 	g_assert (data != NULL);
-	slot = (VirtualSlot*)data;
-	
-	g_return_if_fail (GCK_IS_MANAGER (slot->session_manager));
-	
+	apt = (Apartment*)data;
+
+	g_return_if_fail (GCK_IS_MANAGER (apt->session_manager));
+
 	/* Unreference all the sessions */
-	for (l = slot->sessions; l; l = g_list_next (l)) {
-		
+	for (l = apt->sessions; l; l = g_list_next (l)) {
+
 		/* Some sanity checks to make sure things have remained as expected */
 		g_return_if_fail (GCK_IS_SESSION (l->data));
-		g_return_if_fail (gck_session_get_slot_id (l->data) == slot->slot_id);
-		g_return_if_fail (gck_session_get_manager (l->data) == slot->session_manager);
-		g_return_if_fail (gck_session_get_logged_in (l->data) == slot->logged_in);
-		
+		g_return_if_fail (gck_session_get_apartment (l->data) == apt->apt_id);
+		g_return_if_fail (gck_session_get_manager (l->data) == apt->session_manager);
+		g_return_if_fail (gck_session_get_logged_in (l->data) == apt->logged_in);
+
 		g_object_unref (l->data);
 	}
-	
-	g_list_free (slot->sessions);
-	g_object_unref (slot->session_manager);
-	
-	g_slice_free (VirtualSlot, slot);
+
+	g_list_free (apt->sessions);
+	g_object_unref (apt->session_manager);
+
+	g_slice_free (Apartment, apt);
 }
 
-static VirtualSlot*
-virtual_slot_new (GckModuleClass *klass, CK_SLOT_ID slot_id)
+static Apartment*
+apartment_new (GckModuleClass *klass, CK_SLOT_ID slot_id, CK_G_APPLICATION_PTR app)
 {
-	VirtualSlot *slot;
+	Apartment *apt;
 
-	slot = g_slice_new0 (VirtualSlot);
-	slot->session_manager = g_object_new (klass->manager_type, "for-token", FALSE, NULL);
-	slot->logged_in = FALSE;
-	slot->sessions = NULL;
-	slot->slot_id = slot_id;
-	
-	return slot;
+	apt = g_slice_new0 (Apartment);
+	apt->session_manager = g_object_new (GCK_TYPE_MANAGER, "for-token", FALSE, NULL);
+	apt->logged_in = CKU_NONE;
+	apt->sessions = NULL;
+	apt->slot_id = slot_id;
+
+	if (app) {
+		if (!app->applicationId)
+			app->applicationId = gck_util_next_handle () << 8;
+		apt->app_id = app->applicationId;
+		apt->app_ptr = app;
+	} else {
+		apt->app_id = 0;
+		apt->app_ptr = NULL;
+	}
+
+	apt->apt_id = APARTMENT_ID (apt->slot_id, apt->app_id);
+
+	return apt;
 }
 
-static VirtualSlot*
-lookup_virtual_slot (GckModule *self, CK_SLOT_ID slot_id)
+static Apartment*
+lookup_apartment (GckModule *self, CK_ULONG apartment)
 {
 	g_assert (GCK_IS_MODULE (self));
-	return g_hash_table_lookup (self->pv->virtual_slots_by_id, &slot_id);
+	return g_hash_table_lookup (self->pv->apartments_by_id, &apartment);
 }
 
 static void
-register_virtual_slot (GckModule *self, VirtualSlot *slot)
+register_apartment (GckModule *self, Apartment *apt)
 {
-	g_assert (slot);
+	g_assert (apt);
 	g_assert (GCK_IS_MODULE (self));
-	g_assert (!g_hash_table_lookup (self->pv->virtual_slots_by_id, &(slot->slot_id)));
-	
-	g_hash_table_insert (self->pv->virtual_slots_by_id, 
-	                            gck_util_ulong_alloc (slot->slot_id), slot);
+	g_assert (!g_hash_table_lookup (self->pv->apartments_by_id, &(apt->apt_id)));
+
+	g_hash_table_insert (self->pv->apartments_by_id,
+	                     gck_util_ulong_alloc (apt->apt_id), apt);
 }
 
 static void
-unregister_virtual_slot (GckModule *self, VirtualSlot *slot)
+unregister_apartment (GckModule *self, Apartment *apt)
 {
-	g_assert (slot);
+	g_assert (apt);
 	g_assert (GCK_IS_MODULE (self));
-	
-	if (!g_hash_table_remove (self->pv->virtual_slots_by_id, &(slot->slot_id)))
+
+	if (!g_hash_table_remove (self->pv->apartments_by_id, &(apt->apt_id)))
 		g_assert_not_reached ();
 }
 
+static void
+mark_login_apartment (GckModule *self, Apartment *apt, CK_USER_TYPE user)
+{
+	GList *l;
+
+	g_assert (apt);
+	g_assert (GCK_IS_MODULE (self));
+
+	/* Mark all sessions in the partition as logged in */
+	for (l = apt->sessions; l; l = g_list_next (l))
+		gck_session_set_logged_in (l->data, user);
+	apt->logged_in = user;
+}
+
 static void
 parse_argument (GckModule *self, char *arg)
 {
@@ -242,8 +327,9 @@ parse_argument (GckModule *self, char *arg)
 		*(value++) = 0;
 
 	g_strstrip (arg);
-	g_strstrip (value);
-	
+	if (value)
+		g_strstrip (value);
+
 	g_return_if_fail (GCK_MODULE_GET_CLASS (self)->parse_argument);
 	GCK_MODULE_GET_CLASS (self)->parse_argument (self, arg, value);
 }
@@ -318,10 +404,84 @@ done:
 	g_free (dup);
 }
 
+
+static gboolean
+complete_transient_remove (GckTransaction *transaction, GckModule *self, GckObject *object)
+{
+	if (gck_transaction_get_failed (transaction))
+		add_transient_object (self, NULL, object);
+	g_object_unref (object);
+	return TRUE;
+}
+
+static void
+remove_transient_object (GckModule *self, GckTransaction *transaction, GckObject *object)
+{
+	g_assert (GCK_IS_MODULE (self));
+	g_assert (GCK_IS_OBJECT (object));
+
+	g_object_ref (object);
+
+	gck_object_expose (object, FALSE);
+	if (!g_hash_table_remove (self->pv->transient_objects, object))
+		g_return_if_reached ();
+	g_object_set (object, "store", NULL, NULL);
+
+	if (transaction) {
+		gck_transaction_add (transaction, self,
+		                     (GckTransactionFunc)complete_transient_remove, 
+		                     g_object_ref (object));
+	}
+
+	g_object_unref (object);
+}
+
+static gboolean
+complete_transient_add (GckTransaction *transaction, GckModule *self, GckObject *object)
+{
+	if (gck_transaction_get_failed (transaction))
+		remove_transient_object (self, NULL, object);
+	g_object_unref (object);
+	return TRUE;
+}
+
+static void
+add_transient_object (GckModule *self, GckTransaction *transaction, GckObject *object)
+{
+	g_assert (GCK_IS_MODULE (self));
+	g_assert (GCK_IS_OBJECT (object));
+
+	/* Must not already be associated with a session or manager */
+	g_return_if_fail (gck_object_get_manager (object) == self->pv->token_manager);
+	g_return_if_fail (g_hash_table_lookup (self->pv->transient_objects, object) == NULL);
+
+	g_hash_table_insert (self->pv->transient_objects, object, g_object_ref (object));
+	g_object_set (object, "store", self->pv->transient_store, NULL);
+	gck_object_expose (object, TRUE);
+
+	if (transaction) {
+		gck_transaction_add (transaction, self,
+		                     (GckTransactionFunc)complete_transient_add, 
+		                     g_object_ref (object));
+	}
+}
+
 /* -----------------------------------------------------------------------------
  * OBJECT 
  */
 
+static const CK_SLOT_INFO* 
+gck_module_real_get_slot_info (GckModule *self)
+{
+	return &default_slot_info;
+}
+
+static const CK_TOKEN_INFO*
+gck_module_real_get_token_info (GckModule *self)
+{
+	return &default_token_info;
+}
+
 static void 
 gck_module_real_parse_argument (GckModule *self, const gchar *name, const gchar *value)
 {
@@ -357,68 +517,87 @@ gck_module_real_login_change (GckModule *self, CK_SLOT_ID slot_id, CK_UTF8CHAR_P
 }
 
 static CK_RV
-gck_module_real_login_user (GckModule *self, CK_SLOT_ID slot_id, CK_UTF8CHAR_PTR pin, CK_ULONG n_pin)
+gck_module_real_login_user (GckModule *self, CK_ULONG apartment, CK_UTF8CHAR_PTR pin, CK_ULONG n_pin)
 {
-	VirtualSlot *slot;
-	GList *l;
-	
-	slot = lookup_virtual_slot (self, slot_id);
-	g_return_val_if_fail (slot, CKR_GENERAL_ERROR);
+	Apartment *apt;
 
-	/* Mark all sessions in the partition as logged in */
-	for (l = slot->sessions; l; l = g_list_next (l)) 
-		gck_session_set_logged_in (l->data, TRUE);
-	slot->logged_in = TRUE;
-	
+	apt = lookup_apartment (self, apartment);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
+
+	mark_login_apartment (self, apt, CKU_USER);
 	return CKR_OK;
 }
 
 static CK_RV
-gck_module_real_logout_user (GckModule *self, CK_SLOT_ID slot_id)
+gck_module_real_login_so (GckModule *self, CK_ULONG apartment, CK_UTF8CHAR_PTR pin, CK_ULONG n_pin)
 {
-	VirtualSlot *slot;
-	GList *l;
+	Apartment *apt;
 
-	/* Calculate the partition identifier */
-	slot = lookup_virtual_slot (self, slot_id);
-	g_return_val_if_fail (slot, CKR_GENERAL_ERROR);
+	apt = lookup_apartment (self, apartment);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
 
-	/* Mark all sessions in the partition as logged out */
-	for (l = slot->sessions; l; l = g_list_next (l)) 
-		gck_session_set_logged_in (l->data, FALSE);
-	slot->logged_in = FALSE;
-	
-	/* Derived classes should override if they want actual login */
+	mark_login_apartment (self, apt, CKU_SO);
 	return CKR_OK;
 }
 
-static GObject* 
+static CK_RV
+gck_module_real_logout_any (GckModule *self, CK_ULONG apartment)
+{
+	Apartment *apt;
+
+	/* Calculate the partition identifier */
+	apt = lookup_apartment (self, apartment);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
+
+	mark_login_apartment (self, apt, CKU_NONE);
+	return CKR_OK;
+}
+
+static GObject*
 gck_module_constructor (GType type, guint n_props, GObjectConstructParam *props) 
 {
 	GckModule *self = GCK_MODULE (G_OBJECT_CLASS (gck_module_parent_class)->constructor(type, n_props, props));
+	CK_ATTRIBUTE attr;
+
 	g_return_val_if_fail (self, NULL);	
 
+	/* Register store attributes */
+	attr.type = CKA_LABEL;
+	attr.pValue = "";
+	attr.ulValueLen = 0;
+	gck_store_register_schema (self->pv->transient_store, &attr, NULL, 0);
 
-	
 	return G_OBJECT (self);
 }
 
 static void
 gck_module_init (GckModule *self)
 {
+	gck_timer_initialize ();
+
 	self->pv = G_TYPE_INSTANCE_GET_PRIVATE (self, GCK_TYPE_MODULE, GckModulePrivate);
-	self->pv->token_manager = g_object_new (GCK_MODULE_GET_CLASS (self)->manager_type, "for-token", TRUE, NULL);
-	self->pv->sessions_by_handle = g_hash_table_new_full (gck_util_ulong_hash, gck_util_ulong_equal, 
+	self->pv->token_manager = g_object_new (GCK_TYPE_MANAGER, "for-token", TRUE, NULL);
+	self->pv->sessions_by_handle = g_hash_table_new_full (gck_util_ulong_hash, gck_util_ulong_equal,
 	                                                      gck_util_ulong_free, g_object_unref);
-	self->pv->virtual_slots_by_id = g_hash_table_new_full (gck_util_ulong_hash, gck_util_ulong_equal, 
-	                                                       gck_util_ulong_free, virtual_slot_free);
-	self->pv->factories = g_array_new (FALSE, TRUE, sizeof (GckFactoryInfo));
-	
-	g_atomic_int_set (&(self->pv->handle_counter), 1);
-	
+	self->pv->apartments_by_id = g_hash_table_new_full (gck_util_ulong_hash, gck_util_ulong_equal,
+	                                                    gck_util_ulong_free, apartment_free);
+	self->pv->factories = g_array_new (FALSE, TRUE, sizeof (GckFactory));
+
+	self->pv->handle_counter = 1;
+
+	/* Create the store for transient objects */
+	self->pv->transient_store = GCK_STORE (gck_memory_store_new ());
+	self->pv->transient_objects = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, gck_util_dispose_unref);
+
 	/* Register session object factories */
-	gck_module_register_factory (self, GCK_FACTORY_PRIVATE_KEY);
-	gck_module_register_factory (self, GCK_FACTORY_PUBLIC_KEY);
+	gck_module_register_factory (self, GCK_FACTORY_AES_KEY);
+	gck_module_register_factory (self, GCK_FACTORY_CERTIFICATE);
+	gck_module_register_factory (self, GCK_FACTORY_CREDENTIAL);
+	gck_module_register_factory (self, GCK_FACTORY_NULL_KEY);
+	gck_module_register_factory (self, GCK_FACTORY_DH_PRIVATE_KEY);
+	gck_module_register_factory (self, GCK_FACTORY_PRIVATE_XSA_KEY);
+	gck_module_register_factory (self, GCK_FACTORY_DH_PUBLIC_KEY);
+	gck_module_register_factory (self, GCK_FACTORY_PUBLIC_XSA_KEY);
 }
 
 static void
@@ -426,15 +605,16 @@ gck_module_dispose (GObject *obj)
 {
 	GckModule *self = GCK_MODULE (obj);
 
+	g_hash_table_remove_all (self->pv->transient_objects);
+	g_hash_table_remove_all (self->pv->sessions_by_handle);
+	g_hash_table_remove_all (self->pv->apartments_by_id);
+
 	if (self->pv->token_manager)
 		g_object_unref (self->pv->token_manager);
 	self->pv->token_manager = NULL;
-	
-	g_hash_table_remove_all (self->pv->virtual_slots_by_id);
-	g_hash_table_remove_all (self->pv->sessions_by_handle);
-	
+
 	g_array_set_size (self->pv->factories, 0);
-    
+
 	G_OBJECT_CLASS (gck_module_parent_class)->dispose (obj);
 }
 
@@ -442,13 +622,19 @@ static void
 gck_module_finalize (GObject *obj)
 {
 	GckModule *self = GCK_MODULE (obj);
+
+	g_hash_table_destroy (self->pv->transient_objects);
+	self->pv->transient_objects = NULL;
 	
+	g_object_unref (self->pv->transient_store);
+	self->pv->transient_store = NULL;
+
 	g_assert (self->pv->token_manager == NULL);
 
-	g_assert (g_hash_table_size (self->pv->virtual_slots_by_id) == 0);
-	g_hash_table_destroy (self->pv->virtual_slots_by_id);
-	self->pv->virtual_slots_by_id = NULL;
-	
+	g_assert (g_hash_table_size (self->pv->apartments_by_id) == 0);
+	g_hash_table_destroy (self->pv->apartments_by_id);
+	self->pv->apartments_by_id = NULL;
+
 	g_assert (g_hash_table_size (self->pv->sessions_by_handle) == 0);
 	g_hash_table_destroy (self->pv->sessions_by_handle);
 	self->pv->sessions_by_handle = NULL;
@@ -456,6 +642,8 @@ gck_module_finalize (GObject *obj)
 	g_array_free (self->pv->factories, TRUE);
 	self->pv->factories = NULL;
 
+	gck_timer_shutdown ();
+
 	G_OBJECT_CLASS (gck_module_parent_class)->finalize (obj);
 }
 
@@ -472,6 +660,10 @@ gck_module_set_property (GObject *obj, guint prop_id, const GValue *value,
 		if (args != NULL && args->pReserved != NULL) 
 			parse_arguments (self, args->pReserved);
 		break;
+	case PROP_MUTEX:
+		self->pv->mutex = g_value_get_pointer (value);
+		g_return_if_fail (self->pv->mutex);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
 		break;
@@ -511,19 +703,17 @@ gck_module_class_init (GckModuleClass *klass)
 	gobject_class->set_property = gck_module_set_property;
 	gobject_class->get_property = gck_module_get_property;
     
-	klass->module_info = &default_module_info;
-	klass->slot_info = &default_slot_info;
-	klass->token_info = &default_token_info;
-	klass->manager_type = GCK_TYPE_MANAGER;
-	klass->session_type = GCK_TYPE_SESSION;
-	
+	klass->get_slot_info = gck_module_real_get_slot_info;
+	klass->get_token_info = gck_module_real_get_token_info;
 	klass->parse_argument = gck_module_real_parse_argument;
 	klass->refresh_token = gck_module_real_refresh_token;
 	klass->store_token_object = gck_module_real_store_token_object;
 	klass->remove_token_object = gck_module_real_remove_token_object;
 	klass->login_change = gck_module_real_login_change;
 	klass->login_user = gck_module_real_login_user;
-	klass->logout_user = gck_module_real_logout_user;
+	klass->logout_user = gck_module_real_logout_any;
+	klass->login_so = gck_module_real_login_so;
+	klass->logout_so = gck_module_real_logout_any;
 	
 	g_object_class_install_property (gobject_class, PROP_MANAGER,
 	           g_param_spec_object ("manager", "Manager", "Token object manager", 
@@ -536,6 +726,10 @@ gck_module_class_init (GckModuleClass *klass)
 	g_object_class_install_property (gobject_class, PROP_INITIALIZE_ARGS,
 	           g_param_spec_pointer ("initialize-args", "Initialize Args", "Arguments passed to C_Initialize", 
 	                                 G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
+
+	g_object_class_install_property (gobject_class, PROP_MUTEX,
+	           g_param_spec_pointer ("mutex", "Mutex", "Module mutex",
+	                                 G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
 }
 
 /* -----------------------------------------------------------------------------
@@ -553,12 +747,17 @@ gck_module_get_manager (GckModule *self)
 gboolean
 gck_module_get_write_protected (GckModule *self)
 {
-	g_return_val_if_fail (GCK_IS_MODULE (self), FALSE);
-	g_return_val_if_fail (GCK_MODULE_GET_CLASS (self)->token_info, FALSE);
-	return (GCK_MODULE_GET_CLASS (self)->token_info->flags & CKF_WRITE_PROTECTED) ? TRUE : FALSE;
+	const CK_TOKEN_INFO* info;
+	
+	g_return_val_if_fail (GCK_IS_MODULE (self), TRUE);
+	g_return_val_if_fail (GCK_MODULE_GET_CLASS (self)->get_token_info, TRUE);
+	
+	info = (GCK_MODULE_GET_CLASS (self)->get_token_info) (self);
+	g_return_val_if_fail (info, TRUE);
+	
+	return info->flags & CKF_WRITE_PROTECTED;
 }
 
-
 GckSession*
 gck_module_lookup_session (GckModule *self, CK_SESSION_HANDLE handle)
 {
@@ -599,6 +798,22 @@ gck_module_logout_user (GckModule *self, CK_SLOT_ID slot_id)
 	return GCK_MODULE_GET_CLASS (self)->logout_user (self, slot_id);	
 }
 
+CK_RV
+gck_module_login_so (GckModule *self, CK_SLOT_ID slot_id, CK_UTF8CHAR_PTR pin, CK_ULONG n_pin)
+{
+	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_GENERAL_ERROR);
+	g_assert (GCK_MODULE_GET_CLASS (self)->login_so);
+	return GCK_MODULE_GET_CLASS (self)->login_so (self, slot_id, pin, n_pin);
+}
+
+CK_RV
+gck_module_logout_so (GckModule *self, CK_SLOT_ID slot_id)
+{
+	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_GENERAL_ERROR);
+	g_assert (GCK_MODULE_GET_CLASS (self)->logout_so);
+	return GCK_MODULE_GET_CLASS (self)->logout_so (self, slot_id);	
+}
+
 CK_ULONG
 gck_module_next_handle (GckModule *self)
 {
@@ -624,7 +839,11 @@ gck_module_store_token_object (GckModule *self, GckTransaction *transaction, Gck
 	g_return_if_fail (GCK_IS_MODULE (self));
 	g_return_if_fail (GCK_IS_OBJECT (object));
 	g_assert (GCK_MODULE_GET_CLASS (self)->store_token_object);
-	GCK_MODULE_GET_CLASS (self)->store_token_object (self, transaction, object);
+
+	if (gck_object_is_transient (object))
+		add_transient_object (self, transaction, object);
+	else
+		GCK_MODULE_GET_CLASS (self)->store_token_object (self, transaction, object);
 }
 
 void
@@ -633,25 +852,29 @@ gck_module_remove_token_object (GckModule *self, GckTransaction *transaction, Gc
 	g_return_if_fail (GCK_IS_MODULE (self));
 	g_return_if_fail (GCK_IS_OBJECT (object));
 	g_assert (GCK_MODULE_GET_CLASS (self)->remove_token_object);
-	GCK_MODULE_GET_CLASS (self)->remove_token_object (self, transaction, object);
+
+	if (gck_object_is_transient (object))
+		remove_transient_object (self, transaction, object);
+	else
+		GCK_MODULE_GET_CLASS (self)->remove_token_object (self, transaction, object);
 }
 
 void
-gck_module_register_factory (GckModule *self, GckFactoryInfo *factory)
+gck_module_register_factory (GckModule *self, GckFactory *factory)
 {
 	g_return_if_fail (GCK_IS_MODULE (self));
 	g_return_if_fail (factory);
 	g_return_if_fail (factory->attrs || !factory->n_attrs);
-	g_return_if_fail (factory->factory);
+	g_return_if_fail (factory->func);
 	
 	g_array_append_val (self->pv->factories, *factory);
 	self->pv->factories_sorted = FALSE;
 }
 
-GckFactory
+GckFactory*
 gck_module_find_factory (GckModule *self, CK_ATTRIBUTE_PTR attrs, CK_ULONG n_attrs)
 {
-	GckFactoryInfo *factory;
+	GckFactory *factory;
 	gboolean matched;
 	gulong j;
 	gsize i;
@@ -665,7 +888,7 @@ gck_module_find_factory (GckModule *self, CK_ATTRIBUTE_PTR attrs, CK_ULONG n_att
 	}
 	
 	for (i = 0; i < self->pv->factories->len; ++i) {
-		factory = &(g_array_index (self->pv->factories, GckFactoryInfo, i));
+		factory = &(g_array_index (self->pv->factories, GckFactory, i));
 		
 		matched = TRUE;
 		for (j = 0; j < factory->n_attrs; ++j) {
@@ -676,12 +899,24 @@ gck_module_find_factory (GckModule *self, CK_ATTRIBUTE_PTR attrs, CK_ULONG n_att
 		}
 		
 		if (matched)
-			return factory->factory;
+			return factory;
 	}
 	
 	return NULL;
 }
 
+/*
+ * Hidden method to get the mutex for a module. This is for timers to be
+ * able to reenter the module. Don't use this method.
+ */
+
+GMutex*
+_gck_module_get_scary_mutex_that_you_should_not_touch (GckModule *self)
+{
+	g_return_val_if_fail (GCK_IS_MODULE (self), NULL);
+	return self->pv->mutex;
+}
+
 /* -----------------------------------------------------------------------------
  * PKCS#11
  */
@@ -699,7 +934,7 @@ gck_module_C_GetInfo (GckModule *self, CK_INFO_PTR info)
 	klass = GCK_MODULE_GET_CLASS (self);
 	g_return_val_if_fail (klass, CKR_GENERAL_ERROR);
 	
-	memcpy (info, klass->module_info, sizeof (CK_INFO));
+	memcpy (info, &default_module_info, sizeof (CK_INFO));
 	
 	/* Extend all the strings appropriately */
 	extend_space_string (info->libraryDescription, sizeof (info->libraryDescription));
@@ -729,9 +964,9 @@ gck_module_C_GetSlotList (GckModule *self, CK_BBOOL token_present, CK_SLOT_ID_PT
 	}
 	
 	g_return_val_if_fail (slot_list, CKR_ARGUMENTS_BAD);
-	
+
 	/* Answer C_GetSlotList with 0 for app */
-	slot_list[0] = CK_GNOME_MAKE_APPARTMENT (GCK_SLOT_ID, 0);
+	slot_list[0] = GCK_SLOT_ID;
 	*count = 1;
 	return CKR_OK;
 }
@@ -739,21 +974,26 @@ gck_module_C_GetSlotList (GckModule *self, CK_BBOOL token_present, CK_SLOT_ID_PT
 CK_RV
 gck_module_C_GetSlotInfo (GckModule *self, CK_SLOT_ID id, CK_SLOT_INFO_PTR info)
 {
+	const CK_SLOT_INFO *original;
 	GckModuleClass *klass;
 	
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
-	if (CK_GNOME_APPARTMENT_SLOT (id) != GCK_SLOT_ID)
+
+	if (id != GCK_SLOT_ID)
 		return CKR_SLOT_ID_INVALID;
 	if (info == NULL)
 		return CKR_ARGUMENTS_BAD;
-	
+
 	/* Any slot ID is valid for partitioned module */
 	
 	klass = GCK_MODULE_GET_CLASS (self);
 	g_return_val_if_fail (klass, CKR_GENERAL_ERROR);
+	g_return_val_if_fail (klass->get_slot_info, CKR_GENERAL_ERROR);
 	
-	memcpy (info, klass->slot_info, sizeof (CK_SLOT_INFO));
+	original = (klass->get_slot_info) (self);
+	g_return_val_if_fail (original, CKR_GENERAL_ERROR);
+	
+	memcpy (info, original, sizeof (CK_SLOT_INFO));
 	
 	/* Extend all the strings appropriately */
 	extend_space_string (info->manufacturerID, sizeof (info->manufacturerID));
@@ -765,32 +1005,34 @@ gck_module_C_GetSlotInfo (GckModule *self, CK_SLOT_ID id, CK_SLOT_INFO_PTR info)
 CK_RV
 gck_module_C_GetTokenInfo (GckModule *self, CK_SLOT_ID id, CK_TOKEN_INFO_PTR info)
 {
+	const CK_TOKEN_INFO *original;
 	GckModuleClass *klass;
 	
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
-	if (CK_GNOME_APPARTMENT_SLOT (id) != GCK_SLOT_ID)
+
+	if (id != GCK_SLOT_ID)
 		return CKR_SLOT_ID_INVALID;
 	if (info == NULL)
 		return CKR_ARGUMENTS_BAD;
-	
+
 	/* Any slot ID is valid for partitioned module */
 	
 	klass = GCK_MODULE_GET_CLASS (self);
 	g_return_val_if_fail (klass, CKR_GENERAL_ERROR);
+	g_return_val_if_fail (klass->get_token_info, CKR_GENERAL_ERROR);
 	
-	memcpy (info, klass->token_info, sizeof (CK_TOKEN_INFO));
+	original = (klass->get_token_info) (self);
+	g_return_val_if_fail (original, CKR_GENERAL_ERROR);
 	
+	memcpy (info, original, sizeof (CK_TOKEN_INFO));
+
 	/* Extend all the strings appropriately */
 	extend_space_string (info->label, sizeof (info->label));
 	extend_space_string (info->manufacturerID, sizeof (info->manufacturerID));
 	extend_space_string (info->model, sizeof (info->model));
 	extend_space_string (info->serialNumber, sizeof (info->serialNumber));
-	
-	/* We don't purport to have a clock */
-	memset (info->utcTime, 0, sizeof (info->utcTime));
-	
-	return CKR_OK;	
+
+	return CKR_OK;
 }
 
 CK_RV
@@ -801,12 +1043,12 @@ gck_module_C_GetMechanismList (GckModule *self, CK_SLOT_ID id,
 	guint i;
 	
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
-	if (CK_GNOME_APPARTMENT_SLOT (id) != GCK_SLOT_ID)
+
+	if (id != GCK_SLOT_ID)
 		return CKR_SLOT_ID_INVALID;
 	if (count == NULL)
 		return CKR_ARGUMENTS_BAD;
-	
+
 	/* Just want to get the count */
 	if (mech_list == NULL) {
 		*count = n_mechanisms;
@@ -834,8 +1076,8 @@ gck_module_C_GetMechanismInfo (GckModule *self, CK_SLOT_ID id,
 	guint index;
 	
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
-	if (CK_GNOME_APPARTMENT_SLOT (id) != GCK_SLOT_ID)
+
+	if (id != GCK_SLOT_ID)
 		return CKR_SLOT_ID_INVALID;
 	if (info == NULL)
 		return CKR_ARGUMENTS_BAD;
@@ -863,39 +1105,58 @@ CK_RV
 gck_module_C_OpenSession (GckModule *self, CK_SLOT_ID id, CK_FLAGS flags, CK_VOID_PTR user_data, 
                           CK_NOTIFY callback, CK_SESSION_HANDLE_PTR result)
 {
+	CK_G_APPLICATION_PTR app;
 	CK_SESSION_HANDLE handle;
-	VirtualSlot *slot;
 	gboolean read_only;
 	GckSession *session;
-	
+	Apartment *apt = NULL;
+
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
-	if (CK_GNOME_APPARTMENT_SLOT (id) != GCK_SLOT_ID)
+
+	if (APARTMENT_SLOT (id) != GCK_SLOT_ID)
 		return CKR_SLOT_ID_INVALID;
 	if (!result)
 		return CKR_ARGUMENTS_BAD;
-	
+
 	if (!(flags & CKF_SERIAL_SESSION))
 		return CKR_SESSION_PARALLEL_NOT_SUPPORTED;
 
-	/* Lookup or register the virtual slot */
-	slot = lookup_virtual_slot (self, id);
-	if (slot == NULL) {
-		slot = virtual_slot_new (GCK_MODULE_GET_CLASS (self), id);
-		register_virtual_slot (self, slot);
+	/*
+	 * If they're calling us with the 'application' extension, then
+	 * allocate or use our application identifier.
+	 */
+	if (flags & CKF_G_APPLICATION_SESSION) {
+		app = user_data;
+		if (app == NULL)
+			return CKR_ARGUMENTS_BAD;
+		if (app->applicationId)
+			apt = lookup_apartment (self, APARTMENT_ID (id, app->applicationId));
+	} else {
+		app = NULL;
+		apt = lookup_apartment (self, APARTMENT_ID (id, 0));
 	}
 
+	/* The first time this application is accessing, or closed all sessions, allocate new */
+	if (apt == NULL) {
+		apt = apartment_new (GCK_MODULE_GET_CLASS (self), id, app);
+		register_apartment (self, apt);
+	}
+
+	/* Can't open read only session if SO login */
+	if (apt->logged_in == CKU_SO && !(flags & CKF_RW_SESSION))
+		return CKR_SESSION_READ_WRITE_SO_EXISTS;
+
 	/* Make and register a new session */
 	handle = gck_module_next_handle (self);
 	read_only = !(flags & CKF_RW_SESSION);
-	session = g_object_new (GCK_MODULE_GET_CLASS (self)->session_type, "slot-id", slot->slot_id, 
-	                        "read-only", read_only, "handle", handle, "module", self, 
-	                        "manager", slot->session_manager, "logged-in", slot->logged_in, NULL);
-	slot->sessions = g_list_prepend (slot->sessions, session);
-	
+	session = g_object_new (GCK_TYPE_SESSION, "slot-id", apt->slot_id, "apartment", apt->apt_id,
+	                        "read-only", read_only, "handle", handle, "module", self,
+	                        "manager", apt->session_manager, "logged-in", apt->logged_in, NULL);
+	apt->sessions = g_list_prepend (apt->sessions, session);
+
 	/* Track the session by handle */
-	g_hash_table_insert (self->pv->sessions_by_handle, 
-	                     gck_util_ulong_alloc (handle), 
+	g_hash_table_insert (self->pv->sessions_by_handle,
+	                     gck_util_ulong_alloc (handle),
 	                     g_object_ref (session));
 	
 	*result = handle;
@@ -906,67 +1167,85 @@ CK_RV
 gck_module_C_CloseSession (GckModule *self, CK_SESSION_HANDLE handle)
 {
 	GckSession *session;
-	CK_SLOT_ID slot_id;
-	VirtualSlot *slot;
+	CK_ULONG apt_id;
+	Apartment *apt;
 	GList *link;
-	
+
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
+
 	session = gck_module_lookup_session (self, handle);
 	if (session == NULL)
 		return CKR_SESSION_HANDLE_INVALID;
 
 	/* Calculate the virtual slot */
-	slot_id = gck_session_get_slot_id (session);
-	slot = lookup_virtual_slot (self, slot_id);
-	g_return_val_if_fail (slot, CKR_GENERAL_ERROR);
+	apt_id = gck_session_get_apartment (session);
+	apt = lookup_apartment (self, apt_id);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
 
-	link = g_list_find (slot->sessions, session);
+	link = g_list_find (apt->sessions, session);
 	g_return_val_if_fail (link, CKR_GENERAL_ERROR);
-	slot->sessions = g_list_delete_link (slot->sessions, link);
+	apt->sessions = g_list_delete_link (apt->sessions, link);
 	g_object_unref (session);
-	if (!slot->sessions) 
-		unregister_virtual_slot (self, slot);
+	if (!apt->sessions)
+		unregister_apartment (self, apt);
 
 	if (!g_hash_table_remove (self->pv->sessions_by_handle, &handle))
 		g_assert_not_reached ();
-	
+
 	return CKR_OK;
 }
 
 CK_RV
 gck_module_C_CloseAllSessions (GckModule *self, CK_SLOT_ID id)
 {
-	VirtualSlot *slot;
+	Apartment *apt;
 	CK_SESSION_HANDLE handle;
 	GList *l;
-	
+
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
-	if (CK_GNOME_APPARTMENT_SLOT (id) != GCK_SLOT_ID)
+
+	if (APARTMENT_SLOT (id) != GCK_SLOT_ID)
 		return CKR_SLOT_ID_INVALID;
 
-	/* Calculate the virtual slot */
-	slot = lookup_virtual_slot (self, id);
-	if (slot == NULL)
+	apt = lookup_apartment (self, id);
+	if (apt == NULL)
 		return CKR_OK;
-	
+
 	/* Unregister all its sessions */
-	for (l = slot->sessions; l; l = g_list_next (l)) {
+	for (l = apt->sessions; l; l = g_list_next (l)) {
 		handle = gck_session_get_handle (l->data);
 		if (!g_hash_table_remove (self->pv->sessions_by_handle, &handle))
 			g_assert_not_reached ();
 	}
 
-	unregister_virtual_slot (self, slot);
-	return CKR_OK;	
+	unregister_apartment (self, apt);
+	return CKR_OK;
 }
 
 CK_RV
 gck_module_C_InitPIN (GckModule* self, CK_SESSION_HANDLE handle, 
-                      CK_UTF8CHAR_PTR pin, CK_ULONG pin_len)
+                      CK_UTF8CHAR_PTR pin, CK_ULONG n_pin)
 {
-	return CKR_FUNCTION_NOT_SUPPORTED;
+	GckSession *session;
+	Apartment *apt;
+	CK_ULONG apt_id;
+
+	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
+
+	session = gck_module_lookup_session (self, handle);
+	if (session == NULL)
+		return CKR_SESSION_HANDLE_INVALID;
+
+	/* Calculate the virtual slot */
+	apt_id = gck_session_get_apartment (session);
+	apt = lookup_apartment (self, apt_id);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
+
+	if (apt->logged_in != CKU_SO)
+		return CKR_USER_NOT_LOGGED_IN;
+
+	/* Our InitPIN assumes an uninitialized PIN */
+	return gck_module_login_change (self, apt_id, NULL, 0, pin, n_pin);
 }
 
 CK_RV
@@ -974,33 +1253,34 @@ gck_module_C_SetPIN (GckModule* self, CK_SESSION_HANDLE handle, CK_UTF8CHAR_PTR
                      CK_ULONG old_pin_len, CK_UTF8CHAR_PTR new_pin, CK_ULONG new_pin_len)
 {
 	GckSession *session;
-	VirtualSlot *slot;
-	CK_SLOT_ID slot_id;
-	
+	Apartment *apt;
+	CK_ULONG apt_id;
+
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
+
 	session = gck_module_lookup_session (self, handle);
 	if (session == NULL)
 		return CKR_SESSION_HANDLE_INVALID;
 
 	/* Calculate the virtual slot */
-	slot_id = gck_session_get_slot_id (session);
-	slot = lookup_virtual_slot (self, slot_id);
-	g_return_val_if_fail (slot, CKR_GENERAL_ERROR);
+	apt_id = gck_session_get_apartment (session);
+	apt = lookup_apartment (self, apt_id);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
 
-	return gck_module_login_change (self, slot_id, old_pin, old_pin_len, new_pin, new_pin_len);
+	return gck_module_login_change (self, apt_id, old_pin, old_pin_len, new_pin, new_pin_len);
 }
 
 CK_RV
 gck_module_C_Login (GckModule *self, CK_SESSION_HANDLE handle, CK_USER_TYPE user_type,
                     CK_UTF8CHAR_PTR pin, CK_ULONG pin_len)
 {
-	CK_SLOT_ID slot_id;
+	CK_ULONG apt_id;
 	GckSession *session;
-	VirtualSlot *slot;
-	
+	Apartment *apt;
+	GList *l;
+
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
+
 	session = gck_module_lookup_session (self, handle);
 	if (session == NULL)
 		return CKR_SESSION_HANDLE_INVALID;
@@ -1009,44 +1289,62 @@ gck_module_C_Login (GckModule *self, CK_SESSION_HANDLE handle, CK_USER_TYPE user
 	if (user_type == CKU_CONTEXT_SPECIFIC)
 		return gck_session_login_context_specific (session, pin, pin_len);
 
-	/* We don't have support for SO logins */
-	if (user_type == CKU_SO) 
-		return CKR_USER_TYPE_INVALID;
-	
 	/* Some random crap... */
-	if (user_type != CKU_USER)
+	if (user_type != CKU_USER && user_type != CKU_SO)
 		return CKR_USER_TYPE_INVALID;
 
 	/* Calculate the virtual slot */
-	slot_id = gck_session_get_slot_id (session);
-	slot = lookup_virtual_slot (self, slot_id);
-	g_return_val_if_fail (slot, CKR_GENERAL_ERROR);
+	apt_id = gck_session_get_apartment (session);
+	apt = lookup_apartment (self, apt_id);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
 
-	if (slot->logged_in)
+	if (apt->logged_in != CKU_NONE)
 		return CKR_USER_ALREADY_LOGGED_IN;
-	
-	return gck_module_login_user (self, slot_id, pin, pin_len);
+
+	if (user_type == CKU_SO) {
+
+		/* Can't login as SO if read-only sessions exist */
+		for (l = apt->sessions; l; l = g_list_next (l)) {
+			if (gck_session_get_read_only (l->data))
+				return CKR_SESSION_READ_ONLY_EXISTS;
+		}
+
+		return gck_module_login_so (self, apt_id, pin, pin_len);
+
+	} else if (user_type == CKU_USER) {
+		return gck_module_login_user (self, apt_id, pin, pin_len);
+
+	} else {
+		return CKR_USER_TYPE_INVALID;
+	}
 }
 
 CK_RV
 gck_module_C_Logout (GckModule *self, CK_SESSION_HANDLE handle)
 {
-	CK_SLOT_ID slot_id;
-	VirtualSlot *slot;
+	CK_ULONG apt_id;
+	Apartment *apt;
 	GckSession *session;
-	
+
 	g_return_val_if_fail (GCK_IS_MODULE (self), CKR_CRYPTOKI_NOT_INITIALIZED);
-	
+
 	session = gck_module_lookup_session (self, handle);
 	if (session == NULL)
 		return CKR_SESSION_HANDLE_INVALID;
 
-	slot_id = gck_session_get_slot_id (session);
-	slot = lookup_virtual_slot (self, slot_id);
-	g_return_val_if_fail (slot, CKR_GENERAL_ERROR);
+	apt_id = gck_session_get_apartment (session);
+	apt = lookup_apartment (self, apt_id);
+	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
 
-	if (!slot->logged_in)
+	if (apt->logged_in == CKU_NONE)
 		return CKR_USER_NOT_LOGGED_IN;
 
-	return gck_module_logout_user (self, slot_id);
+	else if (apt->logged_in == CKU_USER)
+		return gck_module_logout_user (self, apt_id);
+
+	else if (apt->logged_in == CKU_SO)
+		return gck_module_logout_so (self, apt_id);
+
+	else
+		g_return_val_if_reached (CKR_GENERAL_ERROR);
 }
